/*
 * MinMaxMin.cpp
 *
 *  Created on: 21-set-2009
 *      Author: gio
 */

#include <JTC/jtc.h>
#include "barrier.h"


#define N 4

/*
 * Semplice classe che memorizza gli indici di potenziali candidati ad essere il valore finale
 * di MinMaxMin
 */

class Candidati : public JTCMonitor

{
	int* candidati;
	int numCandidati;
public:
	Candidati (int M)
	{
		candidati = new int[M];
		numCandidati = 0;
	}
	void addCandidato(int i)
	{
		JTCSynchronized sync(*this);
		candidati[numCandidati++] = i;
	}
	int getCandidato(int i)
	{
		JTCSynchronized sync(*this);
        return candidati[i];
	}

	int size()
	{
		JTCSynchronized sync(*this);
        return numCandidati;
	}
	~Candidati()
	{
		delete candidati;
	}

};

/*
 * Il Worker di tipo A prende in input un pezzo di vettore, e riempie il vettore di candidati v con tutti
 * gli indici pari al valoreCercato.
 */

class WorkerA : public JTCThread
{
	int* A;
    int S1,S2;
	Candidati* v;
	Barrier* b;
	int valoreCercato;

public:
	WorkerA (int* A, int S1, int S2, Candidati* v, Barrier* b, int valoreCercato)
	{
		this->A = A;
	    this->S1 = S1;
	    this->S2 = S2;
	    this->v = v;
	    this->b = b;
	    this->valoreCercato = valoreCercato;
	}
	virtual void run()
	{
		for(int i = S1; i < S2; i++)
		{
			if (A[i] == valoreCercato)
				v->addCandidato(i);
		}
		b->done();
	}

};

/*
 * Il worker di tipo B prende in input un pezzo di vettore,
 * cerca il valoreCercato in esso, ma solo negli indici
 * che compaiono in c.
 */

class WorkerB : public JTCThread
{
	int* B;
    int S1,S2;
    Candidati* c;
	Candidati* v;
	Barrier* b;
	int valoreCercato;

public:
	WorkerB (int* B, int S1, int S2, Candidati* c, Candidati* v, Barrier* b, int valoreCercato)
	{
		this->B = B;
	    this->S1 = S1;
	    this->S2 = S2;
        this->v = v;
        this->c = c;
	    this->b = b;
	    this->valoreCercato = valoreCercato;
	}

	virtual void run()
	{
		for(int i = S1; i < S2; i++)
		{
			int cand;
			if (B[cand = c->getCandidato(i)] == valoreCercato)
				v->addCandidato(cand);
		}
		b->done();
	}

};

void cercaTutti(int* A, int M, Candidati* c, Barrier* b, int valoreCercato)

{
	   WorkerA* T[N];
	   int S = M / N;
	   int* A1 = A;
	   int i;
	   for (i = 0; i < N-1; i++)
	   {
		   T[i] = new WorkerA(A1,i*S,(i+1)*S,c,b,valoreCercato);
		   T[i]->run();
		   A1 += S;
	   }
	   T[N-1] = new WorkerA(A1,i*S, M,c,b,valoreCercato);
	   T[N-1]->run();

	   b->waitForCompletion();
	   b->reset();

}

void cercaSelezionati(int* V, Candidati* c, Candidati* v, Barrier* b, int valoreCercato)

{
	WorkerB* W[N];
	int S = c->size() / N;
	int i;
	for (i = 0; i < N-1; i++)
	{
		   W[i] = new WorkerB(V,i*S,(i+1)*S,c,v,b,valoreCercato);
		   W[i]->run();
	}
	W[N-1] = new WorkerB(V,i*S, c->size(),c,v,b,valoreCercato);
	W[N-1]->run();

	b->waitForCompletion();
	b->reset();
}

int MinMaxMin(int* A, int* B, int* C, int M)

{
   Candidati vincitoriA(M);
   Barrier b(N);

   /*
    *  Fase 1: ricerca sul vettore A.
    */

   cercaTutti(A,M,&vincitoriA,&b,0);

   if (vincitoriA.size() == 1)
	   return vincitoriA.getCandidato(0);

   /*
    * Fase 2: non essendoci un unico elemento di A che vale 0, vuol dire che o A contiene tutti 1 oppure contiene pi di
    * uno 0. Si procede a esaminare B.
    */
   Candidati vincitoriB(vincitoriA.size() > 0 ? vincitoriA.size() : M);
   if (vincitoriA.size() > 0)
   {
       cercaSelezionati(B,&vincitoriA,&vincitoriB,&b,1);
   }
   else{
	   cercaTutti(B,M,&vincitoriB,&b,1);
   }


   if (vincitoriB.size() == 1)
	   return vincitoriB.getCandidato(0);

   /*
     * Fase 3: non essendoci un unico elemento tra i candidati che vale 1, vuol dire che o vincitoriA
     * contiene tutti 0 oppure contiene pi di
     * un 1. Si procede a esaminare C.
     */
      Candidati vincitoriC(vincitoriB.size() > 0 ? vincitoriB.size() : M);
      if (vincitoriB.size() > 0)
      {
          cercaSelezionati(C,&vincitoriB,&vincitoriC,&b,0);
      }
      else
      {
    	   cercaTutti(C,M,&vincitoriC,&b,0);
      }

   if (vincitoriC.size() == 1)
   	   return vincitoriC.getCandidato(0);

   return -1;
}

int A[] = {1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0};
int B[] = {0,0,0,0,1,1,1,0,1,0,1,0,1,0,1,0,1};
int C[] = {0,0,0,0,1,0,1,1,1,1,1,1,1,1,1,1,1};

#include <iostream>

using namespace std;

int main()
{
    JTCInitialize init;
    cout << MinMaxMin(A,B,C,17);
}
